/*******************************************************************************
 * Program...: greenhouse.ino Monitor environment in new greenhouse
 * Author....: Jan-marie Newton
 * Date......: 2023-10-03
 * History...: 
 * Notes.....: version 3.5
 *******************************************************************************/
#include <EEPROM.h>
#include <VariableTimedAction.h>
#include <SPI.h>
#include <Ethernet.h>
#include <EthernetUdp.h> // UDP library from: bjoern@cs.stanford.edu 12/30/2008
#include <DHT_Async.h>
#include <Wire.h>


// Constants - set relay pin numbers:
//------------------------------------------------------------------------------
#define Uint  unsigned int
#define Ulong unsigned long
#define RELAYS     2  // number of active relays
#define HVAC_RELAY 5
#define PUMP_RELAY 6
#define FAN_RELAY 14
#define LEVEL      7
#define DHT11_PIN  8   // internal control unit tempurature & humidity
#define DHT22_PIN  9   // Greenhouse internal tempurature & humidity
#define INTERVAL   15000ul // best if its multiple of 60000 ie 60 secs

// DHT_SENSOR_TYPEs: DHT_TYPE_11 | DHT_TYPE_21 | DHT_TYPE_22
DHT_Async dht_11(DHT11_PIN, DHT_TYPE_11);
DHT_Async dht_22(DHT22_PIN, DHT_TYPE_22);

// Initialize Variables
//------------------------------------------------------------------------------
EthernetUDP Udp;         // An EthernetUDP instance to let us send and receive packets over UDP
IPAddress ip(192, 168, 1, 110);
byte mac[]             = {0xDE, 0xAD, 0xBE, 0xEF, 0x1E, 0x9B}; // Enter a MAC address and IP address
unsigned int localPort = 8888;      // local port to listen on
// buffers for receiving and sending data
char packetBuffer[UDP_TX_PACKET_MAX_SIZE];           // buffer to hold incoming packet
char replyBuffer[]     = "O.000.40.400.50.0.727.55.00000008 ";  // a string to send back
const char* cmds[]     = {"Heat", "Cool", "Off", "SP", "on", "off", "Water", "Stop", "Reset", "HVAC"};

//----------------------------
bool Debug   = true;
float ih     = 50;
float it     = 40;
float gh     = 50;
float gt     = 40;
float io     = -2.5; // box temperature measurement error offset
Uint theHour = 23;


//******************************************************
// Object.: Status updater
// Notes..: this-> is the objects pointer to itself
//******************************************************
class StatusReply : public VariableTimedAction {

private:
   int relayState[RELAYS]; // record of reach relay status
   int schedule[8];        // watering schedule
   int set_Point = 78;
   int heatPoint = 40;
   char mode     = 'O';    // O = Off H = Heat C = Cool
   Ulong rain    = 0;      // how long to keep the water pump on
   Uint dow      = 0;      // day of week
   int lvl       = 0;      // status of water take float switch
   int fan       = 0;      // system cooling fan

   //updates the data and reply buffer
   //-----------------------------------------------------------
   unsigned long run() {
      getStatus();
      return 0;
   }
   //---------------------------------------------------------------
  
public:

   //------------------------------------
   StatusReply() {
      for(int i=0; i<7; i++) {schedule[i] = 0;}
      schedule[7]   = 8;
      for(int i=0; i<RELAYS; i++) {relayState[i] = LOW;}
      start(INTERVAL); 
   }
   //-----------------------------------------------------------
   void getStatus() {
      if (rain > 0) rain -= min(rain,INTERVAL);

      lvl = digitalRead(LEVEL);      // 0 = has water  1 = empty
      if ((lvl == 1 || rain == 0) && digitalRead(PUMP_RELAY) == HIGH)  {
         digitalWrite(PUMP_RELAY, LOW);
         relayState[1] = LOW;
      } else if (lvl == 0 && rain > 0 && relayState[1] == LOW) {
         digitalWrite(PUMP_RELAY, HIGH);
         relayState[1] = HIGH;
      }

      relayState[0] = digitalRead(HVAC_RELAY);
      switch (mode) {
        case 'O' :
           if (relayState[0] == HIGH) {
              digitalWrite(HVAC_RELAY, LOW);
              relayState[0] = LOW;
              digitalWrite(FAN_RELAY, LOW);
              fan = LOW;
           }
           break;

        case 'H' :
           if (gt < heatPoint && relayState[0] == LOW) {
              relayState[0] = HIGH;
              digitalWrite(HVAC_RELAY, HIGH);
           } else if (gt > heatPoint && relayState[0] == HIGH) {
              relayState[0] = LOW;
              digitalWrite(HVAC_RELAY, LOW);
           }
           break;

        case 'C' :
           if (gt > set_Point && relayState[0] == LOW) {
              relayState[0] = HIGH;
              digitalWrite(HVAC_RELAY, HIGH);
              digitalWrite(FAN_RELAY, HIGH);
              fan = HIGH;
           } else if (gt < set_Point && relayState[0] == HIGH) {
              relayState[0] = LOW;
              digitalWrite(HVAC_RELAY, LOW);
              digitalWrite(FAN_RELAY, LOW);
              fan = LOW;
           }
      }

      if (Debug) {
         Serial.print("Mode:");
         Serial.print(mode);
         Serial.print(" Relay:");
         Serial.print(relayState[0]);
         Serial.print(relayState[1]);
         Serial.print(fan); 
         Serial.print(" SetPoint:");
         Serial.print((mode == 'C' ? set_Point : heatPoint));
         Serial.print(" Temp:");
         Serial.print(gt,1);     // greenhouse temperature
         Serial.print(" ");
         Serial.print(gh,0);     // greenhouse humidity
         Serial.print("%  Water:");
         Serial.print(1-lvl);
         Serial.print(" Rain:");
         Serial.print(rain != 0);
         Serial.print(" Hour:");
         Serial.print(theHour);
         Serial.print("=");
         Serial.print(schedule[7]);
         Serial.print(" CPU:");
         Serial.print(it,1);     // inside temperature
         Serial.print(" ");
         Serial.print(ih,0);     // inside humidity
         Serial.print("% ");
         for(int i=0; i < 7; i++) { Serial.print((char)(48 + schedule[i])); }
         Serial.println("");
      }

      String rec;
      rec += mode; rec += "\t";                 // thermostat mode
      for(int i=0; i < RELAYS; i++)
         rec += (relayState[i]);                // current relay dispostion
      rec += fan;

      rec += "\t";
      rec += (mode == 'C' ? set_Point : heatPoint); rec += "\t";
      rec += (int)round(gt * 10); rec += "\t";  // greenhouse temperature to 1/10 of degree
      rec += (int)gh; rec += "\t";              // greenhouse humidity
      rec += (1-lvl); rec += "\t";              // low water float switch
      rec += (int)round(it * 10); rec += "\t";  // inside temperature to 1/10 of degree
      rec += (int)ih; rec += "\t";              // inside humidity
      for(int i=0; i<=7; i++)
         rec += schedule[i];

      rec.toCharArray(replyBuffer, rec.length() + 1);
   }
   //------------------------------------
   void waterViaSchedule() {
      if (theHour != schedule[7] || mode == 'O') return;
      if (packetBuffer[13] >= '0' && packetBuffer[13] <= '6')
         dow = packetBuffer[13] - 48;

      setRain( schedule[dow] );
      if (rain > 1 && lvl == 0) {  // 0 = has water  1 = empty
         relayState[1] = HIGH;
         digitalWrite(PUMP_RELAY, HIGH);
      }
      getStatus();
   }
   //------------------------------------
   void setMode(char m) {
      if (mode == m)
         return;

      mode = m;
      EEPROM.update(0,m);
      for(int i=0; i<7; i++) {schedule[i] = 0;}
      if (mode == 'H') {
         //schedule[0] = 1;   // [0,0,0,1,0,0,0,8]
         schedule[3] = 1;
      } else {
         schedule[0] = 1;   // [1,0,1,0,0,1,0,8]
         schedule[2] = 1;
         schedule[5] = 1;
      }
   }
   //------------------------------------
   void setRain(Ulong r=1) {
      // plus INTERVAL cause getStatus will immediately decrease rain
      rain = (r * 60000) + INTERVAL;
   }
   //------------------------------------
   // Waterdd:m dd=dow 0=sunday, m=mins 0-9  Water:hr:H  hr = 0-23
   //------------------------------------
   void setSchedule() {
      packetBuffer[7] = 0;
      int d = max(0, atoi( &packetBuffer[5] ));
      if (packetBuffer[8] == 'h' || packetBuffer[8] == 'H')
         schedule[7] = (d <= 23) ? d : 7;
      else if (d <= 6)
         schedule[d] = min(9, atoi( &packetBuffer[8] ));
   }
   //------------------------------------
   void setPumpOn() {
      if (lvl == 1) return; // There is no water
      setRain();
      if (relayState[1] == LOW)
         relayState[1] = HIGH;

      if (digitalRead(PUMP_RELAY) == LOW)
         digitalWrite(PUMP_RELAY, HIGH);
   }
   //------------------------------------
   void setPumpOff() {
      setRain(0);
      if (relayState[1] == HIGH)
         relayState[1] = LOW;

      if (digitalRead(PUMP_RELAY) == HIGH)
         digitalWrite(PUMP_RELAY, LOW);
   }
   //------------------------------------
   void setPoint(int sp) {
      if (sp < 0 || sp > 80 || mode == 'O') return;

      if (mode == 'H')
        heatPoint = sp;

      else if (mode == 'C')
        set_Point = sp;

      //char t = sp;
      //EEPROM.update(1,t);
   }
   //------------------------------------
   void setHVAC() {
      relayState[0] = 1 - digitalRead(HVAC_RELAY);
      fan           = relayState[0];
      digitalWrite(HVAC_RELAY, relayState[0]);
      digitalWrite(FAN_RELAY, fan);
   }
   //------------------------------------
} status;
//**********************************************************


//----------------------------------------------------------------------
static bool measure_environment() {
   static unsigned long dht11_timestamp = millis();
   static unsigned long dht22_timestamp = millis();
   int knt = 0;

   /* Measure once every four seconds. */
   if (millis() - dht11_timestamp > 3950ul) {
       if (dht_11.measure(&it, &ih)) {
          dht11_timestamp = millis();
          it = dht_11.convertCtoF(it + io);
          ++knt;
       }
   }
   if (millis() - dht22_timestamp > 3950ul) {
       if (dht_22.measure(&gt, &gh)) {
          dht22_timestamp = millis();
          gt = dht_22.convertCtoF(gt);
          ++knt;
       }
   }
   return (knt == 2);
}
//----------------------------------------------------------------------


// Setup Initialization
//---------------------------------------------------------------------------
void setup() {
   //Initialize serial and wait for port to open:
   if (Debug) {
      Serial.begin(115200);
      while (!Serial) { ; } // wait for serial port to connect.
      Serial.println("Program uploaded complete!");
      Serial.println("");
   }

   // start the Ethernet and UDP:
   Ethernet.begin(mac, ip);
   Udp.begin(localPort);

   // initialize the relay pins as an output:
   pinMode(PUMP_RELAY, OUTPUT);
   pinMode(HVAC_RELAY, OUTPUT);
   pinMode(FAN_RELAY,  OUTPUT);
   pinMode(LEVEL, INPUT_PULLUP);
   digitalWrite(PUMP_RELAY, LOW);
   digitalWrite(HVAC_RELAY, LOW);

   // get mode from eeprom memory
   char e = EEPROM.read(0);
   if (e == 'O' or e == 'C' or e == 'H') {
      status.setMode(e);
      //e = EEPROM.read(1);
      //status.setPoint(e);
   }
}
//------------------------------------------------------------------------------

// Main event loop
//------------------------------------------------------------------------------
void loop() {
   VariableTimedAction::updateActions();
   // Measure temperature and humidity.
   measure_environment();

   // if there's data available, read a packet
   int packetSize = Udp.parsePacket();
   if (packetSize) {
      Udp.read(packetBuffer, UDP_TX_PACKET_MAX_SIZE); // read the packet into packetBuffer
         /*Serial.print("Received packet of size ");
         Serial.println(packetSize);
         Serial.print("From ");
         IPAddress remote = Udp.remoteIP();
         for (int i = 0; i < 4; i++) {
              Serial.print(remote[i], DEC);
              if (i < 3) Serial.print(".");
          }
          Serial.print(", port ");
          Serial.println(Udp.remotePort());
          Serial.print("Contents: ");
          Serial.println(packetBuffer);*/

      if (packetBuffer[0] == 'r') {
         // expecting: ryymmddHHMMSSw (where w=dow 0=sunday)
         packetBuffer[9]  = 0;
         int hr = max(0, atoi(&packetBuffer[7]));
         if (hr != theHour && hr < 24) {
            theHour = hr;
            status.waterViaSchedule();
         }
      } else {
         for (int i=0; i < 10; i++)
            obeyCommmad(cmds[i]);
      }
      // send a reply to the IP address and port that sent us the packet we received
      Udp.beginPacket(Udp.remoteIP(), Udp.remotePort());
      Udp.write(replyBuffer);
      Udp.endPacket();
   }
}
//------------------------------------------------------------------------------

// Process incoming command
//------------------------------------------------------------------------------
void obeyCommmad(const char* cmd) {
   char *ret = strstr(packetBuffer, cmd);
   if (ret == NULL) return;

   //Serial.print("set command: ");  Serial.println(cmd);
   
   if (cmd == "SP")
      status.setPoint(atoi( &packetBuffer[2] ));

   else if (cmd == "Heat" || cmd == "Cool" || cmd == "Off")
      status.setMode(cmd[0]);

   else if (cmd == "on")
      status.setPumpOn();

   else if (cmd == "off")
      status.setPumpOff();

   else if (cmd == "Water")
      status.setSchedule();

   else if (cmd == "HVAC")
      status.setHVAC();

   else if (cmd == "Reset") {
      status.setPumpOff();
      status.setMode('O');
   }

   status.getStatus();
}
//------------------------------------------------------------------------------
